/* This file was generated by upb_generator from the input file:
 *
 *     envoy/config/core/v3/grpc_service.proto
 *
 * Do not edit -- your changes will be discarded when the file is
 * regenerated. */

#ifndef ENVOY_CONFIG_CORE_V3_GRPC_SERVICE_PROTO_UPB_MINITABLE_H_
#define ENVOY_CONFIG_CORE_V3_GRPC_SERVICE_PROTO_UPB_MINITABLE_H_

#include "upb/generated_code_support.h"

// Must be last.
#include "upb/port/def.inc"

#ifdef __cplusplus
extern "C" {
#endif

extern const upb_MiniTable envoy__config__core__v3__GrpcService_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__EnvoyGrpc_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc__SslCredentials_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc__GoogleLocalCredentials_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc__ChannelCredentials_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc__CallCredentials_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc__CallCredentials__ServiceAccountJWTAccessCredentials_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc__CallCredentials__GoogleIAMCredentials_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc__CallCredentials__MetadataCredentialsFromPlugin_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc__CallCredentials__StsService_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc__ChannelArgs_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc__ChannelArgs__Value_msg_init;
extern const upb_MiniTable envoy__config__core__v3__GrpcService__GoogleGrpc__ChannelArgs__ArgsEntry_msg_init;

extern const upb_MiniTableFile envoy_config_core_v3_grpc_service_proto_upb_file_layout;

#ifdef __cplusplus
}  /* extern "C" */
#endif

#include "upb/port/undef.inc"

#endif  /* ENVOY_CONFIG_CORE_V3_GRPC_SERVICE_PROTO_UPB_MINITABLE_H_ */
